/*
 *  Copyright (C) 2010, Davidlohr Bueso <dave@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * Some ideas from Rik van Riel's OOM Killer.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/seq_file.h>
#include <linux/string.h>
#include <linux/time.h>

/* number of processes to be stored and listed */
#define NPROCS 10

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Davidlohr Bueso <dave@gnu.org>");
MODULE_DESCRIPTION("Shows Top 10 proceses using most of the system's memory");

unsigned long get_age(struct timespec ts)
{
	struct timespec uptime;

	/* get the system's uptime */
	do_posix_clock_monotonic_gettime(&uptime);
	monotonic_to_bootbased(&uptime);

	return uptime.tv_sec - ts.tv_sec;
}

struct proc_info {
	pid_t pid;
	char comm[TASK_COMM_LEN];
	unsigned long total_vm;
	unsigned long runtime;
} *proc_info[NPROCS];

static int get_list(void)
{
	struct task_struct *p;
	int j, i=0;

	for_each_process(p) {
		struct mm_struct *mm;
		unsigned long total_vm;

		/*
		 * skip kernel threads and tasks which have already released
		 * their mm.
		 */
		if (!p->mm)
			continue;
		/* skip the init task */
		if (is_global_init(p))
			continue;

		task_lock(p);
		mm = p->mm;
		if (!mm) {
			task_unlock(p);
			return 0;
		}

		total_vm = mm->total_vm;

		task_unlock(p);

		if(i < NPROCS) {
			proc_info[i] = kmalloc(sizeof(struct proc_info), GFP_KERNEL);

			if(!proc_info[i])
				return -ENOMEM;

			proc_info[i]->pid = p->pid;
			proc_info[i]->runtime = get_age(p->real_start_time);
			strncpy(proc_info[i]->comm, p->comm, strlen(p->comm));
			proc_info[i]->total_vm = total_vm;
		}

		else { /* iterate through the list and see if the vm is higher,
			  then replace with new value */
			for(j = 0; j < NPROCS; j++) {
				if(total_vm >= proc_info[j]->total_vm) {
					proc_info[j]->pid = p->pid;
					proc_info[j]->runtime =  get_age(p->real_start_time);
					strncpy(proc_info[j]->comm, p->comm, TASK_COMM_LEN);
					proc_info[j]->total_vm = total_vm;

					break;
				}
			}
		}

		i++;
	}

	return 0;
}

static int highmem_show(struct seq_file *m, void *v)
{
	int i;

	if(!get_list()) {
		for(i = 0; i< NPROCS; i++)
			seq_printf(m, "%s (%d) using: %lu, running for %lu\n",
				   proc_info[i]->comm,
				   proc_info[i]->pid,
				   proc_info[i]->total_vm,
				   (unsigned long) proc_info[i]->runtime);
	}

	return 0;
}

static int highmem_open(struct inode *inode, struct file *file)
{
	return single_open(file, highmem_show, NULL);
}

static const struct file_operations highmem_fops = {
	.owner	= THIS_MODULE,
	.open	= highmem_open,
	.read	= seq_read,
	.llseek	= seq_lseek,
	.release = single_release,
};

int init_module(void)
{
	if(!proc_create("highmem", 0, NULL, &highmem_fops))
		return -ENOMEM;

	return 0;
}
void cleanup_module(void)
{
	int i;

	remove_proc_entry("highmem", NULL);

	for(i=0;i<NPROCS;i++)
		if(proc_info[i])
			kfree(proc_info[i]);

	return;
}
